# runPoseSeq.py
#
#This file provides the function
#
# runPoseSeq(seqfilename,imtemplate)
#
# that loads a file with a sequence of poses described by articulated
# angles, renders them, possibly modifying global location and
# orientation as well as other parameters, and saves images and
# parameters into files.
#
# Note: you need to update the hard-coded directory names appropriately


import poser
scene = poser.Scene()
import sys
sys.path = sys.path + ['c:\\greg\\Poser\\py']
from humanparam import * 
from handmanip import grasp
from scenemanip import *
import os

# the behavior of runPoseSeq depends on the entries in poseSeqOpt.
# This is a global variable; modify before calling runPoseSeq. The values
# below are default values.

poseSeqOpt={}

# probability of changing the clothing in a given frame
poseSeqOpt['probClothesChange']=0;

# if True, always drop the figure to floor (every frame)
poseSeqOpt['drop']=True;

# if True, draw random values for nuisance parameters
poseSeqOpt['nuisance']=False;

# handling orientation. if -1, apply the orientation from the file;
#            if a positive number, that's the orientation in all frames;
#            if a 2-tuple, draw a random orientation between two bounds;
#            if a list, apply the i-th orientation to i-th frame.
poseSeqOpt['orient']=[-1];

# file with mapping of entries of parameter vector to specific angles.
# if None, uses the array humanparams (described in humanparam.py)
poseSeqOpt['paramfile']=None

# saving options:
# if True, save stereo pair (left/right), otherwise just the right image.
poseSeqOpt['doStereo']=False

# if True, save the silhouette for the right image; otherwise no sil.
poseSeqOpt['saveSil']=True

# the # to assign the first pose (affects filenames)
poseSeqOpt['start']=1;

# output directory
poseSeqOpt['outDir']="c:\\greg\\Poser\\test\\"

# saving 3D joint locations. Set to None to prevent saving
poseSeqOpt['saveXYZ']='XYZ.txt';

# units can be 'meter', 'Poser' (1unit=96in), or 'inch'.
poseSeqOpt['units']='meter';

# same for angles
poseSeqOpt['saveAngles']='Angles.txt';

# the 2D location variation of the person. The default is almost fixed
# at origin (the ground plane is XZ).
poseSeqOpt['varyZ']=[-0.01,0.01];
poseSeqOpt['varyX']=[-0.01,0.01];

# number of rendered examples per pose
poseSeqOpt['times']=1;


def runPoseSeq(seqfilename,imtemplate):
    # read a sequence of poses (described by the angle params), and for each
    # pose apply it and render the images.
    # Arguments:
    #   seqfilename : the name of the file with poses
    #   imtemplate : template for frame names (should include %d)
    if poseSeqOpt['paramfile']==None:
        paramlist=map(lambda x: x[0], humanparams);
    else:
        paramlist=readParamList(poseSeqOpt['paramfile']);

    
    if not os.path.exists(poseSeqOpt['outDir']):
        os.makedirs(poseSeqOpt['outDir'])    
    if poseSeqOpt['saveAngles'] != None:
        angf=open(poseSeqOpt['outDir']+poseSeqOpt['saveAngles'],"w");
    if poseSeqOpt['saveXYZ'] != None:
        xyzf=open(poseSeqOpt['outDir']+poseSeqOpt['saveXYZ'],"w");

    # save current pose
    params = captureParam(paramlist);
    
    file=open(seqfilename,"r");
    pose=[None]
    frame=0
    while len(pose)>0:
        pose=readParam(file,paramlist);
        if len(pose)>0:
            # render this body configuration
            for times in range(0,poseSeqOpt['times']):
                
                # apply the pose
                restoreParams(pose);
                hip=scene.Figure("Figure 1").Actor("Hip");
            
                if poseSeqOpt['drop']:
                    scene.Figure("Figure 1").DropToFloor();

                # set orientation
                orient=hip.Parameter('yRotate').Value();
                if len(poseSeqOpt['orient'])==1 and poseSeqOpt['orient'][0] > 0:
                    orient=poseSeqOpt['orient'][0];
                if len(poseSeqOpt['orient'])==2:
                    orient=uniform(poseSeqOpt['orient'][0],poseSeqOpt['orient'][1]);
                if len(poseSeqOpt['orient'])>2:
                    orient=poseSeqOpt['orient'][frame]
                hip.SetParameter('yRotate',orient);
                
                # draw nuisance variables if necessary
                if poseSeqOpt['nuisance']:
                    for p in nuisance_params:
                        setpar(scene.Figure("Figure 1"),p);
                    grasp('l',uniform(-.8,.05))
                    grasp('r',uniform(-.05,.8))

                # save joint locations (in canonical location)
                #hip.SetParameter('xTran',0);
                #hip.SetParameter('zTran',0);
                #scene.DrawAll();
                
                # now set location (in XZ plane)
                locX=uniform(poseSeqOpt['varyX'][0],poseSeqOpt['varyX'][1]);
                locZ=uniform(poseSeqOpt['varyZ'][0],poseSeqOpt['varyZ'][1]);
                hip.SetParameter('xTran',locX);
                hip.SetParameter('zTran',locZ);

                # modify clothing if necessary
                if uniform(0,1) < poseSeqOpt['probClothesChange']:
                    setShirt()
                    setPants()
                    setShoes()
                    setHair()
            
                # render
                eff_frame=frame+poseSeqOpt['start']
                frname=poseSeqOpt['outDir']+imtemplate % eff_frame
                stereoshot(frname,sil=poseSeqOpt['saveSil'],stereo=poseSeqOpt['doStereo']);
                # if required, save the joint locations
                if poseSeqOpt['saveXYZ'] != None:
                    logXYZ(xyzf,units=poseSeqOpt['units'])                

                # if required, save the pose (angles) - indep. of location
                if poseSeqOpt['saveAngles'] != None:
                    writeparam(angf,paramlist)
                frame+=1

            # end of one frame generated for this pose
        # end of this pose
        
    file.close()

    # restore the pose before runPoseSeq.
    restoreParams(params)

    # close files, if opened them
    if poseSeqOpt['saveAngles'] != None:
        angf.close()
    if poseSeqOpt['saveXYZ'] != None:
        xyzf.close()
        
    # return the number of frames rendered
    return frame-poseSeqOpt['start']

# end of runPoseSeq


